package core

import (
	"fmt"

	"git.akyoto.dev/cli/q/src/errors"
	"git.akyoto.dev/cli/q/src/expression"
	"git.akyoto.dev/cli/q/src/token"
	"git.akyoto.dev/cli/q/src/types"
	"git.akyoto.dev/cli/q/src/x86"
)

// CompileCall executes a function call.
// All call registers must hold the correct parameter values before the function invocation.
// Registers that are in use must be saved if they are modified by the function.
// After the function call, they must be restored in reverse order.
func (f *Function) CompileCall(root *expression.Expression) ([]types.Type, error) {
	var (
		pkg      = f.Package
		pkgNode  *expression.Expression
		name     string
		nameNode = root.Children[0]
		fn       *Function
		exists   bool
	)

	if nameNode.IsLeaf() {
		name = nameNode.Token.Text(f.File.Bytes)

		switch name {
		case "len":
			return _len.OutputTypes, f.CompileLen(root)
		case "syscall":
			return nil, f.CompileSyscall(root)
		case "new":
			typ, err := f.CompileNew(root)
			return []types.Type{typ}, err
		case "delete":
			return nil, f.CompileDelete(root)
		case "store":
			return nil, f.CompileMemoryStore(root)
		}
	} else {
		pkgNode = nameNode.Children[0]
		nameNode = nameNode.Children[1]

		pkg = pkgNode.Token.Text(f.File.Bytes)
		name = nameNode.Token.Text(f.File.Bytes)
	}

	if pkg == "kernel32" || pkg == "user32" || pkg == "gdi32" || pkg == "comctl32" {
		parameters := root.Children[1:]
		registers := x86.WindowsInputRegisters[:len(parameters)]

		for i := len(parameters) - 1; i >= 0; i-- {
			_, err := f.ExpressionToRegister(parameters[i], registers[i])

			if err != nil {
				return nil, err
			}
		}

		f.DLLs = f.DLLs.Append(pkg, name)
		f.DLLCall(fmt.Sprintf("%s.%s", pkg, name))
		return nil, nil
	} else if pkg != f.File.Package {
		if f.File.Imports == nil {
			return nil, errors.New(&errors.UnknownPackage{Name: pkg}, f.File, pkgNode.Token.Position)
		}

		imp, exists := f.File.Imports[pkg]

		if !exists {
			return nil, errors.New(&errors.UnknownPackage{Name: pkg}, f.File, pkgNode.Token.Position)
		}

		imp.Used = true
	}

	fn, exists = f.Functions[pkg+"."+name]

	if !exists {
		return nil, errors.New(&errors.UnknownFunction{Name: name}, f.File, nameNode.Token.Position)
	}

	parameters := root.Children[1:]

	if len(parameters) != len(fn.Input) {
		return nil, errors.New(&errors.ParameterCountMismatch{Function: fn.Name, Count: len(parameters), ExpectedCount: len(fn.Input)}, f.File, nameNode.Token.End())
	}

	registers := f.CPU.Input[:len(parameters)]

	for i := len(parameters) - 1; i >= 0; i-- {
		typ, err := f.ExpressionToRegister(parameters[i], registers[i])

		if err != nil {
			return nil, err
		}

		if !types.Is(typ, fn.Input[i].Type) {
			_, expectsPointer := fn.Input[i].Type.(*types.Pointer)

			if expectsPointer && parameters[i].Token.Kind == token.Number && parameters[i].Token.Text(f.File.Bytes) == "0" {
				continue
			}

			return nil, errors.New(&errors.TypeMismatch{Encountered: typ.Name(), Expected: fn.Input[i].Type.Name(), ParameterName: fn.Input[i].Name}, f.File, parameters[i].Token.Position)
		}
	}

	f.CallSafe(fn, registers)
	return fn.OutputTypes, nil
}
